$NOMOD51
$GENONLY
$CONDONLY

$INCLUDE(MG84FL54B.MCU)
$INCLUDE(hid_codes.a51)
$INCLUDE(229_keymap.a51)

DEBOUNCE_COUNT EQU 149 ; equiv to 10mS apparently

POLL_INTERVAL_250 EQU DEBOUNCE_COUNT / 250
POLL_INTERVAL_UNITS EQU DEBOUNCE_COUNT MOD 250

NCOLUMNS EQU 18

;; Data variables
;; ==============
;;
;; Ports:
;;   B1 P3.1 DIP switch 4
;;   B2 P3.2 DIP switch 3
;;   B5 P3.5 Num Lock
;;   B6 P3.6 Caps Lock
;;   B7 P3.7 Scroll Lock
;;   E9 P4.1 DIP switch 2
;;   EA P4.2 DIP switch 1
;;
;; RAM:
;;   00 register bank 0
;;   08 register bank 1
;;   10 register bank 2
;;   18 register bank 3
;;   20 byte ledState;

ledState 	DATA 20h
numLock 	BIT ledState.0
capsLock 	BIT ledState.1
scrollLock 	BIT ledState.2

;;      Used in do_receive(), zeroed in init_vars()
;;      Used in upd_leds() as [? ? ? ? ? Scroll Caps ?] bit mask
;;      Bit 0 used in numlock_on() in connection with NumLock

kbMode 	DATA	21h
;;   21 enum { KBMODE_NORMAL = 0x5A, KBMODE_MACRO = 0xA5 } kb_mode
;;      Used in upd_leds(), assigned in init_vars(), used in decode_fn(), complemented in process_shift_f12()
;;      // num_lock_fn_lock?

shiftF12Count 	DATA 22h
;;   22 byte shift_f12_count;
;;      Zeroed in init_vars(), incremented/tested/zeroed in process_shift_f12()

colIndex	DATA 23h

;;   23 byte column_index;
;;      Zeroed in init_vars(), used in scan_column(), used in same_column_value() as column index,
;;      used in process_key()

colValue	DATA 24h
;;   24 byte column_value;
;;      Assigned column value in scan_column(), used in two_bits_set()

colTemp		DATA 25h
colChanges	DATA 25h
;;   25 union {
;;      byte new_column_value;
;;          Assigned in read_column()
;;      byte old_column_value;
;;          Assigned in scan_column(), used in decode_fn()
;;      byte column_changed;
;;          Assigned in decode_fn()

scanResults	DATA 	26h
;;   26 \ byte key_matrix[18];
;;   37 / Zeroed in init_vars(), used in scan_column(), current/old state of key matrix?

usbTXBusy	DATA	38h

;;   38 bool usb1_transmitting;
;;      Zeroed in init_vars(), zeroed in do_usb_2() when transmission to endpoint 1 is finished,
;;      assigned/waited in process_key()

idleRate	DATA	39h
;;   39 byte idle_rate;
;;      Transmitted in response to vendor request 2, assigned from vendor request 0x0A, never used

protocol	DATA	3Ah
;;      Transmitted in response to vendor request 3, assigned from vendor request 0x0B, never used

keyIndex	DATA	3Bh
;;   3B local in decode_numlock(), decode_fn()

keyCode		DATA	3Ch

;;      assigned/used in decode_fn(), decode_numlock();
;;      assigned in find_macro(); used in store_key()

phantomFlag	DATA	3Dh
;;   3D bool phantom;
;;      Zeroed in init_vars(), scan_column(), used/zeroed in process_key()

normalKey	DATA	3Eh
;;      Zeroed in init_vars(), assigned in decode_fn(), used/zeroed in process_key(), tested in trigger_numlock()
;;      Zeroed in process_key at end of scanning cycle


;;   3F local in find_macro()
;;   40 byte macro_delay_after;
;;      Tested in play_macro(), assigned in find_macro()
;;   41 \ word macro_delay;
;;   42 / Assigned found->items[j].delay in play_macro()
;;   43 byte usb_intr_count;
;;      Decremented in do_usb_2()
;;   44 bool after_macro;
;;      Tested 0x80/zeroed in process_key(); assigned found->items[j].flags in play_macro()
;;   45 bool usb2_transmitting;
;;      Zeroed in init_vars(), zeroed in do_usb_2() when transmission to endpoint 2 is finished

mediaRptReady	DATA 46h
;;   46 bool media_report_ready;
;;      Assigned in decode_fn(), stop_fn(); tested/zeroed in process_key()
;;   47 byte media_keycode;
;;      Zeroed in init_vars(), used in decode_fn() as Fn-pretranslated keycode, zeroed in stop_fn(),
;;      used in processMediaKeycode() as switch
;;   48 \ byte hid_report2[2];
;;   49 / Assigned in processMediaKeycode(), stop_fn(), transmitted in process_key()
;;   4A byte keydown_index;
;;      Zeroed in init_vars(), process_key(); loop count in process_key();
;;      tested/incremented in store_key()

keyModifiers DATA 04bh

;;   4B union {
;;         byte hid_report[8];
;;         struct {
;;             byte key_modifiers;
;;                 bits: [RWin RAlt RShift RCtrl LWin LAlt LShift LCtrl]
;;                 used in process_shift_f12(), process_key()
;;   4C        byte hid_reserved;
;;   4D        byte keys_down[6];
;;   52 }  }
;;      Zeroed in init_vars(), process_key() after transmitting, assigned in process_key().
;;      keys_down[0] == 0xDF in decode_fn() when Fn is down.
;;      keys_down[0] == keys_down[1] == 0 in decode_fn() stops Fn.
;;      keys_down[0], keys_down[1] used in process_shift_f12()
;;      keys_down[0], keys_down[1], keys_down[2] tested/assigned in process_key()
;;   53 local loop counters in two_bits_set(), same_column_value(), decode_fn(),
;;      decode_numlock(), process_key(), find_macro(), play_macro()
;;   54 local loop counters in find_macro(), play_macro()
;;   55 locals in play_macro(), process_key()
;;   56 locals in scan_column(), two_bits_set(), same_column_value(), decode_fn(),
;;      find_macro(), play_macro()
;;   57 \ struct MacroBinding* pBinding
;;   58 / scans memory in find_macro(); corrupted in play_macro() for no reason
;;   59 \ union { MacroData* pMacro; MacroEvent* pMacroEvent; }
;;   5A / Assigned in find_macro() to found->pointer, used in play_macro()

STACK	DATA	5Bh
;;   5B \ hardware stack
;;   7F / 37 bytes
;;   :
;;   BC \ byte clockA[18];
;;   CD /
;;   CE \ byte clockB[18];
;;   DF /
pressedKeys DATA 0E0h
;;   E0 \ struct PressedKey { byte keyindex; byte layer; } pressed_keys[15];
;;   FD /

SRAMTOP		DATA	0FFh

;; XRAM:
;;   0000 byte delay_count;
;;        Decremented by timer0(). timer0() fires every 1000*12 clocks which means 1ms at 12MHz.
;;        initialized and waited for 0 by delay()
;;   0001 byte wdtcr;
;;        Used in timer0(), usb_intr(), keypad()

layerKeysDown	XDATA 03h		; extended for multiple layer keys (pw) and bits changed to layer bits
;;   0003 byte fn_down; // 0x80 if down, 0 if up
;;        Zeroed in init_vars(), assigned/used in process_key()
;;   0004 byte numlock_timer;
;;   0005 byte numlock_state;
;;        Zeroed in int_vars()
;;        Transition from 0x00 to 0xEA in trigger_numlock() if internal numlock is on and a key is pressed
;;        Transition from 0xEA to 0xCC in numlock_on() if internal numlock is on, external off, turning external numlock on
;;        Transition from 0xCC to 0x88 in process_key() if all keys are released, setting timer to 0xFF
;;        Transition from 0x88 to 0xBB in do_delay() when timer expires
;;        Transition from 0xBB to 0x00 in numlock_off(), toggling external numlock [off unless interfered with externally]
;;   0007 byte media_key_down;
;;        Zeroed in init_vars(), used in decode_fn()
;;   000A byte isp_state;
;;        Zeroed in init_isp()
;;        Transition without check from 0x00 to the type of current request in isp_reprogram() called from do_receive()
;;        Value 0x44 in do_transmit() triggers soft reboot
;;        Transition from 0x55 to 0x00 in isp_read_block() sending report 55 AA 00 28 // version check?
;;        In state 0x66 in isp_read_block() read 64 bytes from code/constant memory, prepare for transmitting
;;        Transition from 0x66 to 0x00 in isp_read_block() when isp_count == 0
;;        Transition from any to 0x00 in isp_reprogram() when localX65 >= isp_count resetting localX65 to isp_count (isp_count becomes 0 shortly after)
;;   000B \ byte* isp_ptr;
;;   000C |
;;   000D | Address for ISP read/write operation in isp(), zeroed in init_isp()
;;   000E /
;;   000F \ dword isp_count;
;;   0010 | Count of pages for isp_erase_pages()
;;   0011 | Count of bytes in isp_reprogram() case 0x66 (read macros), 0x77 (erase and write macros)
;;   0012 / Zeroed in init_isp()
;;   0013 local loop counts in isp_reprogram(), isp_erase_pages(), isp_read_block()
;;   
;;	0014 bool remote_wakeup;
;;        Used in get_status(), clear_feature(), set_feature(), do_usb_2(), zeroed in init_hw()
;;   0015 bool configured;
;;        Used in do_usb_2(), assigned in set_configuration(), zeroed in init_hw(),
;;        used in process_key()
;;   0016 byte request_state;
;;        Transition to 0xF0 in do_receive() on overwrite
;;        Transition from 0xF0 to 0x5A on short GET requests
;;        Transition from 0xF0 to 0xA5 on short SET requests
;;        Transition from 0xF0 to 0x55 on set_address request
;;        Transition from 0xF0 to 0x22 on get_macros request
;;        Transition from 0xF0 to 0x33 on set_macros request
;;        Transition from 0x33 to 0xA5 after set_macros done
;;   0017 union {
;;         struct Report report;
;;         struct Request request = {
;;             byte bmRequestType;
;;   0018      byte bRequest;
;;   0019      __low_endian word wValue;
;;   001A
;;   001B      __low_endian word wIndex;
;;   001C
;;   001D      __low_endian word wLength;
;;   001E
;;   :
;;   0056 }}
;;   0057 enum AddressingMode {
;;             RAM_8 = 0,
;;             XRAM_16 = 1,
;;             XRAM_8 = 0xFE,
;;             CMEM_16 // all other values; 0xFF used explicitly
;;        } transmit_addr_mode;
;;   0058 \ byte* to_transmit;
;;   0059 /
;;   005B byte max_transmit;
;;        Used in do_transmit(), assigned 0x40 in init_hw()
;;   005C \ word tx_rx_count
;;   005D / Byte count in do_receive(), vendor_request(), do_transmit(), get_status(),
;;          assigned by get_descriptor(), get_interface()
;;   005E union { byte usb_address; byte usb_configuration; } // minor wtf
;;        Assigned in set_configuration(), set_address request
;;   005F \ word requested_size
;;   0060 / Assigned in get_descriptor()
;;   0063 local working copy of UPCON and UIFLG in do_usb_2()
;;   0064 local counts in do_receive(), do_transmit(); local word-sized response size in get_descriptor()
;;   0065 local copy of addressing mode in receive(), transmit()
;;        local copy of size argument in isp_reprogram()
;;   0066 \ local copy of base address argument in receive(), transmit()
;;   0067 /
;;   023F XRAMTOP
;;
;;
;; Const section:
;;
;;   2800 struct Macros macros = {
;;             byte disabled[0x20];
;;        Bit array, each bit corresponding to a keycode. A 0 indicates that for this keycode
;;        a corresponding array of macros is present
;;   2820      byte binding_count;
;;   2821      MacroBinding bindings[binding_count];
;;   ????      byte data[];
;;        };
;;
;; struct MacroBinding {
;;     byte keycode;   // Structures with same keycode go contiguously
;;     byte modifiers; // [0 0 0 0 0 Alt Shift Ctrl]
;;     MacroData* pItems;
;;     bool delay_after;   // if 0, exits immediately after playing;
;;                         // otherwise, after 100 interrupts or when any key is pressed
;; };
;;
;; struct {
;;     00: byte count;
;;     01: struct MacroEvent {
;;             00: byte hid_report[8];
;;             08: word delay:12;
;;             09: byte flags:4;
;;                 0x80 Send an empty report after this (release all keys)
;;                 0x40 ?
;;                 0x20 ?
;;                 0x10 ?
;;         } items[];
;; } MacroData;
;;

CSEG AT 0000h
    LJMP START

; TIMER 0 INTERRUPT VECTOR
CSEG AT 000Bh
    LJMP DO_TIMER0

; KEYPAD INTERRUPT VECTOR
CSEG AT 006Bh
    LJMP DO_KEYPAD

; USB INTERRUPT VECTOR
CSEG AT 007Bh
    LJMP DO_USB

; Inlude data first to fix address
CSEG AT 100h
$INCLUDE(data.a51)

START:
    MOV R0, #SRAMTOP		; Clear SRAM
    CLR A
clrMem:                                          
    MOV @R0, A              
    DJNZ R0, clrMem         
    MOV SP, #STACK-1
real_start:                                     ;void real_start() {
    LCALL init_interrupts   ;   init_interrupts();
    LCALL init_isp          ;   init_isp();
    LCALL init_vars         ;   init_vars();
    LCALL init_hw            ;   init_hw();
forever:                                          
    LCALL scan_column       ;       scan_column();
    LCALL process_key       ;       process_key();
    LCALL do_delay          ;       do_delay();
    ;LCALL numlock_off       ;       numlock_off();
    SJMP forever             

    
init_interrupts:                      
    CLR A
    MOV IE, A               ;   IE = 0;
    MOV 0ADh, #0A0h         ;   AUXIE = EUSB | EKB;
    MOV 0AEh, #20h          ;   AUXIP = PKBI; // Keypad interrupt
    MOV PSW, A              ;   PSW = 0;
    MOV IP, #12h            ;   IP = PS | PT0; // Serial port | Timer 0
    MOV TMOD, #21h          ;   TMOD = M1 << 4 | M0; // Timer 1: 8-bit auto reload; timer 0: 16-bit
    CLR TR0                 ;   TR0 = false;
    SETB ET0                ;   ET0 = true;
    MOV R0, #1h
    MOV A, 0E1h
    MOVX @R0, A             ;   wdtcr = WDTCR;
    MOVX A, @R0
    ORL A, #37h
    MOVX @R0, A             ;   wdtcr |= ENW | CLRW | PS2 | PS1 | PS0; // enable, clear, prescale 1/256
                            ;     // 32768 * 12 * 256 / 12MHz = 8.39 s
    MOV 0C7h, #58h          ;   CKCON = XCKS3 | XCKS1 | XCKS0; // 12MHz
    ORL 0BFh, #4h           ;   CKCON2.EN_PLL = true;
L0226:                                          ;   do {
    MOV A, 0BFh
    JNB ACC.1, L0226        ;   } while (not CKCON2.PLL_RDY);
    SETB ENABLE_ANY                 ;   EA = true; // enable interrupts
    ORL 0BFh, #8h           ;   CKCON2.EN_USB = true;
    MOV R7, #1h
    LCALL delay             ;   delay(1);
L0228:                                          ;   do {
    MOV DPTR, #ACKCTL
    MOVX A, @DPTR
    ORL A, #1h
    MOVX @DPTR, A           ;       USB_CC.0 = true;
    MOVX A, @DPTR
    JNB ACC.0, L0228        ;   } while (not USB_CC.0);
    MOV R7, #4h
    LCALL delay             ;   delay(4);
L0229:                                          ;   do {
    MOV DPTR, #ACKCTL
    MOVX A, @DPTR
    ORL A, #2h
    MOVX @DPTR, A           ;       USB_CC.1 = true;
    MOVX A, @DPTR
    JNB ACC.1, L0229        ;   } while (not USB_CC.1);
    ORL 0BFh, #20h          ;   CKCON2.OSCDR0 = true;
    
    ; Configure port IO modes
    MOV P1M0, #0FFh         ;  P1 is open drain output (driven low by 0)
    MOV P1M1, #0FFh         ;   
    MOV P2M0, #0FFh         ;  P2 is open drain output
    MOV P2M1, #0FFh         ;  
    MOV P3M0, #18h          ;  P3 bits 3, 4 - open drain output
    MOV P3M1, #18h          ;  P3 bits 0,1,2,5,6,7 bidirectional 
    MOV R7, #0FFh
    LJMP delay              ;   delay(255);
                            ;}

    
init_vars:                                      ;void init_vars() {
    CLR A
    MOV ledState, A              ;   ledState = 0;
    MOV R0, #layerKeysDown
    MOVX @R0, A             ;   fn_down = 0;
    MOV R0, #7h
    MOVX @R0, A             ;   media_key_down = 0;
    MOV R0, #5h
    MOVX @R0, A             ;   numlock_state = 0;
    DEC R0
    MOVX @R0, A             ;   numlock_timer = 0;
    MOV R0, #usbTXBusy
    MOV @R0, A              ;   usb1_transmitting = 0;
    MOV R0, #colIndex
    MOV @R0, A              ;   column_index = 0;
    MOV R0, #normalKey
    MOV @R0, A              ;   normal_key = false;
    DEC R0
    MOV @R0, A              ;   phantom = false;
    MOV R6, A
    MOV R7, #NCOLUMNS
    MOV R5, A
    MOV R3, A
    MOV R2, #0h
    MOV R1, #scanResults
    LCALL memset            ;   memset(&key_matrix, sizeof(key_matrix), 0);
    CLR A
    MOV R0, #4Ah
    MOV @R0, A              ;   keydown_index = 0;
    MOV R6, A
    MOV R7, #8h
    MOV R5, A
    MOV R3, A
    MOV R2, #0h
    MOV R1, #keyModifiers
    LCALL memset            ;   memset(&hid_report, 8, 0);
    MOV R0, #kbMode
    MOV @R0, #5Ah           ;   kb_mode = 0x5A;
    CLR A
    INC R0
    MOV @R0, A              ;   shift_f12_count = 0;
    MOV R0, #45h
    MOV @R0, A              ;   usb2_transmitting = 0;
    MOV R0, #47h
    MOV @R0, A              ;   media_keycode = 0;
    DEC R0
    MOV @R0, A              ;   media_report_ready = 0;
    RET                     ;}

    
leds_off:                                       ;void leds_off() {
    SETB P3.5               ;   P3./NUM = true;
    SETB P3.6               ;   P3./CAPS = true;
    SETB P3.7               ;   P3./SCROLL = true;
    SETB P4.0               ;   P4.0 = true;
    RET                     ;}
    
init_hw:                    ;void init_hw() {
    MOV DPTR, #IEN
    CLR A
    MOVX @DPTR, A           ;       IEN = 0;
    MOV DPTR, #EPINDEX
    MOVX @DPTR, A           ;       EPINDEX = 0;
    MOV DPTR, #EPCON
    MOV A, #5h
    MOVX @DPTR, A           ;       EPCON = RXEPEN | TXEPEN;
    MOVX A, @DPTR
    ORL A, #0C0h
    MOVX @DPTR, A           ;       EPCON.RXSTL = EPCON.TXSTL = true;
    MOV DPTR, #TXCON
    MOV A, #80h
    MOVX @DPTR, A           ;       TXCON = TXCLR;
    MOV DPTR, #RXCON
    MOVX @DPTR, A           ;       RXCON = RXCLR;
    MOV DPTR, #UIE
    MOV A, #97h
    MOVX @DPTR, A           ;       UIE = SOFIE | UTXIE2 | UTXIE1 | URXIE0 | UTXIE0;
    MOV DPTR, #IEN
    MOV A, #6h
    MOVX @DPTR, A           ;       IEN = EFSR | EF;
    MOV R0, #5Bh
    MOV A, #40h
    MOVX @R0, A             ;       max_transmit = 0x40;
    CLR A
    MOV R0, #15h
    MOVX @R0, A             ;       configured = false;
    DEC R0
    MOVX @R0, A             ;       remote_wakeup = 0;
    MOV DPTR, #UPCON
    MOVX A, @DPTR
    ORL A, #80h
    MOVX @DPTR, A           ;       UPCON.CONEN = true;
    RET                     ;}

      
$INCLUDE(isp.a51)
$INCLUDE(usb.a51)
$INCLUDE(macro.a51)
$INCLUDE(keypad.a51)
$INCLUDE(keyboard.a51)


                                                        
END
